/*
	   Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/** @file h3c_math.c
 *  @brief mathematique functions
 *  @author G.Landais
 */

#include "common.h"
#include "h3c_math.h"

/*****************************************************************************/
/** @brief compute coord in cartesian
 *  @param lon right ascension (in degree)
 *  @param lat declination (in degree)
 *  @param vect (OUT) cartesian coordinates
 */
/*****************************************************************************/
void h3c_2xyz(double lon, double lat, _h3c_vector vect)
{
    double coslat = cos(lat*H3C_DEGRA);
    vect[0] = coslat * cos(lon*H3C_DEGRA);
    vect[1] = coslat * sin(lon*H3C_DEGRA);
    vect[2] = sin(lat*H3C_DEGRA);
}

/*****************************************************************************/
/** @brief cartesian to polar
 *  @param in cartesian coordinates
 *  @param out (OUT) polar coordinates
 */
/*****************************************************************************/
void h3c_2radec(double *in, h3c_coord_t *out)
{
    double r2 = in[0]*in[0] + in[1]*in[1];
    out[0] = 0.0;
    if (r2 == 0.0) { /* in case of poles */
        if (in[2] == 0.0) {
                out[0] = 0./0.; out[1] = 0./0.;
            }
        else
                out[1] = (in[2] > 0.0) ? 90.0 : -90.0;
    }
    else {
            out[0] = atan2(in[1], in[0])*H3C_RADEG;
            out[1] = atan2(in[2], sqrt(r2))*H3C_RADEG;
            if (out[0] < 0.0)
                out[0] += 360.0;
    }
}

/*****************************************************************************/
/** @brief vectorial product
 *  @param v1 vector
 *  @param v2 vector
 *  @param r (OUT) resulting vector
 */
/*****************************************************************************/
void h3c_vect_prod(_h3c_vector v1, _h3c_vector v2, _h3c_vector r) {
    r[0] = v1[1]*v2[2] - v1[2]*v2[1];
    r[1] = v1[2]*v2[0] - v1[0]*v2[2];
    r[2] = v1[0]*v2[1] - v1[1]*v2[0];
}

/*****************************************************************************/
/** @brief scalar product
 *  @param v1 vector
 *  @param v2 vector
 *  @return the scalar product
 */
/*****************************************************************************/
double h3c_scalar_product(_h3c_vector v1, _h3c_vector v2) {
    return v1[0]*v2[0] + v1[1]*v2[1] + v1[2]*v2[2];
}

/*****************************************************************************/
/** @brief normalize in the united vector
 *  @param u vector
 */
/*****************************************************************************/
void h3c_normalize(_h3c_vector u) {
    double n = sqrt(u[0]*u[0] + u[1]*u[1] + u[2]*u[2]);
    if (n > 0) {
          u[0] = u[0]/n; u[1] = u[1]/n; u[2] = u[2]/n; 
    }
}

/*****************************************************************************/
/** @brief get the length of a vector
 *  @param v vector
 *  @return the length
 */
/*****************************************************************************/
double h3c_vect_length(_h3c_vector v) {
    return sqrt(v[0]*v[0] + v[1]*v[1] + v[2]*v[2]);
}

/*****************************************************************************/
/** @brief create a unit vector from a z coordinate and an azimuthal angle. 
 *  @param z coordinate
 *  @param phi azimuth
 *  @param res result vector (OUT)
 */
/*****************************************************************************/
void h3c_unit_vector(double z, double phi, _h3c_vector res) {
    double sintheta = sqrt((1.-z)*(1.+z));
    res[0] = sintheta*cos(phi);
    res[1] = sintheta*sin(phi);
    res[2] = z;
}

