/*
	   Copyright (C) 2012 Gilles Landais (CDS)

    Author: Gilles Landais, Strasbourg astronomical Data Center (CDS)
    Email: gilles.landais@unistra.fr

    This file is part of H3C.

    H3C is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    H3C is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with H3C; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/** @file h3c_poly_more.c
 *  @brief other function to deal with polygon
 *  @date 05/08/2011
 *  @author G.Landais
 */

#include "common.h"
#include "h3c_util.h"
#include "h3c_math.h"

#if H3C_BOX
#include "astropos.h"
#endif

struct h3c_polygon {
    int nvertex;
    h3c_coord_t *ra;
    h3c_coord_t *dec;
    _h3c_vector *vect_prod;
    h3c_coord_t center[2];
    h3c_coord_t radius; /* Circle that embedeed the polygon */
    int nside;
    int order;
    int nipix;
    int ninside;
    int nalloc;
    h3c_ipix_t *ipix;
};

/*****************************************************************************/
/** @brief get the center of a polygon (in cartesian coord)
 *  @param n number of vertex
 *  @param in_ra  right ascension of the polygon points
 *  @param in_dec right ascension of the polygon points
 *  @param b (OUT) the center
 */
/*****************************************************************************/
static void h3c_poly_center_xyz(int n,
                                h3c_coord_t *in_ra,
                                h3c_coord_t *in_dec,
                                _h3c_vector b)
{
    int i;
    _h3c_vector v;
    b[0] = b[1] = b[2] = 0.0;
    for (i = 0; i < n; i++) {
            h3c_2xyz(in_ra[i], in_dec[i],v);
        b[0] += v[0];
        b[1] += v[1];
        b[2] += v[2];
    }
    h3c_normalize(b);
}

/*****************************************************************************/
/** @brief get the polgon center
 *  @param n number of points
 *  @param in_ra right ascention of the polygon points
 *  @param in_dec declination ot the polygon points
 *  @param b (OUT) the center
 */
/*****************************************************************************/
void h3c_poly_center(int n, 
                    h3c_coord_t *in_ra,
                    h3c_coord_t *in_dec,
                    h3c_coord_t *b)
{
    _h3c_vector v;
    h3c_poly_center_xyz(n, in_ra, in_dec, v);
    h3c_2radec(v, b);
    /*h3c_log(H3C_DEBUG, "h3c_poly_center: (x,y,z)=(%.3f,%.3f,%.3f)=(%.3f,%.3f)",
                v[0], v[1], v[2], b[0], b[1]);*/
}

/*****************************************************************************/
/** @brief compute the vectoral products of the vertex of a polygon
 *  @param in_ra right ascention of the points of the polygon
 *  @param in_dec declination of the points of the polygon
 *  @pram n number of vertex
 *  @return allocated pointer to the vectorials products
 *
 *  @warn the pointer returneed must be free!
 */
/*****************************************************************************/
static _h3c_vector *h3c_poly_vect_prod (h3c_coord_t *in_ra,
                                        h3c_coord_t *in_dec,
                                        int n) {
    int i;
    _h3c_vector u, v, *vecp, *p, *q = NULL;

    vecp=(_h3c_vector *)malloc(sizeof(_h3c_vector)*n);
    for (i = 0; i < n; i++) {
        if (i > 0) { p = q; q = (p==&u)?&v:&u; } 
        else { p = &u; q = &v; h3c_2xyz(in_ra[0], in_dec[0], *p); }
        h3c_2xyz(in_ra[(i+1)%n], in_dec[(i+1)%n], *q);
        h3c_vect_prod(*p, *q, vecp[i]);
        h3c_normalize(vecp[i]);
    }

    return vecp;
}

/*****************************************************************************/
/** @brief verify if a point is inside a convex polygon
 *  @see h3c_check_sphere_point_in_poly_convex
 */
/*****************************************************************************/
static int _h3c_check_sphere_point_in_poly_convex(int n,
                                   h3c_coord_t *in_ra,
                                   h3c_coord_t *in_dec,
                                   h3c_coord_t ra0,
                                   h3c_coord_t dec0,
                                   _h3c_vector *vecp)
{
    double s; 
    int i,direct;
    _h3c_vector v0, v;

        /* Verify an invalid point ?*/
        if (n<3) return(false);
        h3c_2xyz(ra0, dec0, v0);

        /* compute the centre */
         h3c_poly_center_xyz(n, in_ra, in_dec,v);

        /* compute the orientation */
        s = h3c_scalar_product(v,vecp[0]);
        direct = s>=0;

        for (i = 0; i < n; i++) {
            s = h3c_scalar_product(v0,vecp[i]);
            if (direct) {
                if (s<=0) return H3C_DISJUNCT;
            }
            else {
                if (s>0) return H3C_DISJUNCT;
            }
        }

        return !H3C_DISJUNCT;
}

/*****************************************************************************/
/** @brief verify if a point is inside a convex polygon
 *  @param n number of points
 *  @param in_ra right ascention of the polygon points
 *  @param in_dec declination of the polygon points
 *  @param ra0 right ascention of the point to test
 *  @param dec0 declination of the point to test
 *  @return H3C_DISJUNCT not inside
 *
 *  @warn the convex polygons are just a particular case of this ensemble
 *
 *  This function can be applied for polygon more large than convex.
 *  (see h3c_poly_is_convex)
 *
 *  @note 
 *  - take care of the sens (ie: the order of the point)
 *  - algo from F.Ochsenbein
 */
/*****************************************************************************/
int h3c_check_sphere_point_in_poly_convex(int n,
                                   h3c_coord_t *in_ra,
                                   h3c_coord_t *in_dec,
                                   h3c_coord_t ra0,
                                   h3c_coord_t dec0)
{
    int ret;
    _h3c_vector *vecp;

    /* compute the vectoral products */
    vecp = h3c_poly_vect_prod(in_ra, in_dec, n);

    /* test the point */
    ret = _h3c_check_sphere_point_in_poly_convex(n, in_ra, in_dec,
                                                ra0, dec0, vecp);

    free(vecp);
    return ret;
}

/*****************************************************************************/
/** @brief verify if the polygon is convex
 *  @param n number of points
 *  @param in_ra right ascension of the polygon points
 *  @param in_dec declination of the polygon points
 *  @return 0 no, 1 OK
 *
 *  @warn the convex polygons are just a particular case of this ensemble
 *
 *  In reality the function return true if the function 
 *  h3c_check_sphere_point_in_poly_convex can be applied to verify if
 *  a point is inside the polygon.
 *
 */
/*****************************************************************************/
int h3c_poly_is_convex(int n, h3c_coord_t *in_ra, h3c_coord_t *in_dec)
{
    int ret;
    h3c_coord_t center[2];
    h3c_poly_center(n, in_ra, in_dec, center);
    ret = h3c_check_sphere_point_in_poly_convex(n, in_ra, in_dec,
                                                center[0], center[1]);
    return (ret == H3C_DISJUNCT)?0:1;
}

/*****************************************************************************/
/** @brief get the ipix number in an other order
 *  @param ipix ipix number
 *  @param order_in the order of the given ipix
 *  @param order_out the order of the returned ipix
 *  @return the new ipix number
 */
/*****************************************************************************/
static h3c_ipix_t h3c_ipix2level(h3c_ipix_t ipix, int order_in, int order_out)
{
    int i;
    if (order_in < order_out) {
        for (i = 0; i < order_out-order_in; i++) 
            ipix = ipix<<2;
        return ipix;
    }
    else {
        for (i = 0; i < order_in-order_out; i++)
            ipix = ipix>>2;
        return ipix;
    }
}


/*****************************************************************************/
/** @brief check if a circle intersects a vertex of a polygon
 *  @param polygon polygon structure
 *  @param ra right ascension of the circle
 *  @param dec declination of the circle
 *  @param radius radius of the circle
 *  @return 1 intersect| 0 no
 */
/*****************************************************************************/
static int h3c_sphere_circle_intersect_poly(struct h3c_polygon *polygon,
                                            h3c_coord_t ra,
                                            h3c_coord_t de,
                                            h3c_coord_t radius)
{
    int i;
    _h3c_vector vp;
    h3c_coord_t dist, dist2poly=-1;

    h3c_2xyz(ra, de, vp);

    /* for each vertex ... */
    for (i = 0; i < polygon->nvertex; i++) {
        /* compute the distance */

#if 0
        if (polygon->ra[i] == polygon->ra[(i+1)%polygon->nvertex]) {
             /*dist = (polygon->dec[i])-de;*/
	     if (ra<0) ra+=360;
             dist = (polygon->ra[i])-ra;
        }
        else {
            dist = (H3C_PI_2 - acos(h3c_scalar_product(vp, polygon->vect_prod[i])))*H3C_RADEG; 
        }
#endif
	dist = (H3C_PI_2 - acos(h3c_scalar_product(vp, polygon->vect_prod[i])))*H3C_RADEG;
        if (dist < 0) dist = dist*(-1.);
        if (dist > radius) continue;

#if 1
        /* Refine the list ???
           IF the center of the current ipix (ra,de) is not in the circle
              defined by the center of the polygon with a radius=max(dist(Pi))+r
                (r=the max size of the ipix)
           THEN the ipix is outside the polygon
        */
	if (dist2poly < 0) dist2poly = h3c_dist(ra, de, polygon->center[0], polygon->center[1]);
        if (dist2poly <= (polygon->radius+radius)) return 1 ;
#else
        return 1;
#endif
    }

    return 0;
}

/*****************************************************************************/
/** @brief check if a circle intersects a vertex of a polygon
 *  @param n number of vertex
 *  @param in_ra polygon right ascention
 *  @param in_dec polygon declination
 *  @param ra0 right ascension of the circle
 *  @param dec0 declination of the circle
 *  @param radius radius of the circle
 *  @return 1 intersect| 0 no
 */
/*****************************************************************************/
int h3c_check_sphere_circle_intersect_poly(int n,
                                    h3c_coord_t in_ra[],
                                    h3c_coord_t in_dec[],
                                    h3c_coord_t ra0,
                                    h3c_coord_t dec0,
                                    h3c_coord_t radius)
{
    struct h3c_polygon polygon;
    int intersect;

    polygon.ra = in_ra;
    polygon.dec = in_dec;
    polygon.nvertex = n;

    /* compute the vect. product */
    polygon.vect_prod = h3c_poly_vect_prod(in_ra, in_dec, n);

    intersect = h3c_sphere_circle_intersect_poly(&polygon, ra0, dec0, radius);
    free(polygon.vect_prod);
 
    return intersect;
}

/*****************************************************************************/
/** @brief get the ipix (in upper order) coming from a given ipix and which 
 *         envelop polygon 
 *  @param polygon polygon structure
 *  @param order the order of the given ipix
 *  @param in_ipix the ipix
 *  @return 0 malloc error, 1 OK
 */
/*****************************************************************************/
static int h3c_polygon_ipix_in_upper_order(struct h3c_polygon *polygon, 
                                           int order,
                                           h3c_ipix_t in_ipix)
{
    h3c_ipix_t ipix, ipix_max;
    h3c_coord_t theta, phi, ra, de, radius;
    int nside;

    if (order == polygon->order) {
        if (polygon->nipix > H3C_DEFAULT_MAX_IPIX) return 0;
/*        h3c_log(H3C_DEBUG, "ADD (nside %.f) ipix %ld", pow(2, order), in_ipix); */
        H3C_PUSH(polygon->ipix, polygon->nipix, polygon->nalloc, in_ipix)
        return 1;
    }

    nside = pow(2, order+1);
    radius = h3c_radius(nside);
    ipix = h3c_ipix2level(in_ipix, order, order+1);
    ipix_max = h3c_ipix2level(in_ipix+1, order, order+1);


    for ( ; ipix < ipix_max; ipix++) {
        h3c_pix2ang_nest(nside, ipix, &theta, &phi);

        ra = h3c_phi2ra(phi)*H3C_RADEG;
        de = h3c_theta2dec(theta)*H3C_RADEG;

        /* if center is inside polygon => OK */
        if (_h3c_check_sphere_point_in_poly_convex(polygon->nvertex, 
                                           polygon->ra,
                                           polygon->dec,
                                           ra,
                                           de,
                                           polygon->vect_prod) == 1 ) {
            h3c_ipix_t m, M;
            m = h3c_ipix2level(ipix, order+1, polygon->order);
            M = h3c_ipix2level(ipix+1, order+1, polygon->order);

            if (polygon->nipix+(M-m) > H3C_DEFAULT_MAX_IPIX) return 0;
            for ( ;m < M; m++) {
                H3C_PUSH(polygon->ipix, polygon->nipix, polygon->nalloc, m)
                polygon->ninside++;
            }
            continue;
        }

        /* if circle arroud ipix intersects a vertex */
        if (h3c_sphere_circle_intersect_poly(polygon, ra, de, radius) == 1){
            if (!h3c_polygon_ipix_in_upper_order(polygon, order+1, ipix)) 
                return 0;
        }
    }
    
    return 1;
}

/*****************************************************************************/
/** @brief retrieve the ipix list for a polygon
 *  @param ra right ascention of the points
 *  @param dec declination of the points
 *  @param n number of points
 *  @param nside nside level
 *  @return the list of couple of ipix {[ipix,ipix],[ipix,ipix],....},
 *          NULL if malloc error (see H3C_DEFAULT_MAX_IPIX)
 *
 *  get the ipix where center is inside the polygon
 *  + the ipix which the distance between the center and the vertexes < radius
 *  (where radius is the max ray of circle arroud ipix for a given nside)
 *
 *  P: the center of ipix
 *  vect(OQ) = normalized (vect(OA) ^ vect(OB))
 *  dist = pi/2 - acos(vect(OP).vect(OQ))
 */
/*****************************************************************************/
h3c_ipix_t *h3c_polygon_ipix(h3c_coord_t *in_ra,
                            h3c_coord_t *in_dec,
                            int n,
                            int nside,
                            int *count)
{
#define H3C_MIN_NSIDE 8

    int i,order;
    struct h3c_polygon polygon;
    h3c_ipix_t ipix, max_ipix;
h3c_log(H3C_INFO, "h3c_polygon_ipix nside=%d", nside);

    max_ipix = h3c_npix(H3C_MIN_NSIDE);
    order = h3c_order(H3C_MIN_NSIDE);

    H3C_LIST(polygon.ipix, polygon.nalloc)
    polygon.nipix = 0;
    polygon.ninside = 0;
    polygon.nside = nside;
    polygon.order = h3c_order(nside);
    polygon.ra = in_ra;
    polygon.dec = in_dec;
    polygon.nvertex = n;
    polygon.radius = 0.;

    /* compute the vect. product */
    polygon.vect_prod = h3c_poly_vect_prod(in_ra, in_dec, n);

    /* calculate center and radius of circle that circumscribed the polygon */
    h3c_poly_center(n, in_ra, in_dec, polygon.center);
    for (i=0; i<n; i++) {
        h3c_coord_t r = h3c_dist(polygon.center[0], polygon.center[1],
                                 in_ra[i], in_dec[i]);
        if (polygon.radius < r) polygon.radius = r;
    }

    /* for each ipix (nside = NSIDE_POLYGON_IDX) */
    for (ipix = 0; ipix < max_ipix; ipix++) {
        if (!h3c_polygon_ipix_in_upper_order(&polygon, order, ipix)) {
            h3c_log(H3C_ERROR, "index needs too much ipix (%d) => abort", 
                    H3C_DEFAULT_MAX_IPIX);
            free(polygon.vect_prod);
            free(polygon.ipix);
            return NULL;
        }
    }

    h3c_log(H3C_INFO, "(h3c_polygon_ipix) #ipix=%d (%d inside)", 
            polygon.nipix, polygon.ninside);
    *count = polygon.nipix;
    free(polygon.vect_prod);
    return polygon.ipix;
}

/*****************************************************************************/
/** @brief get the nearest points of a given point which is on a BIG circle
 *  @param v1 cartes. coord. of the cross-product of the 1st BIG circle
 *  @param v2 cartes. coord. of the cross-product of the 2nd BIG circle
 *  @param ra right ascention of the points found (OUT)
 *  @param dec declination of the points found (OUT)
 *  @return 0 the same circle, 1 if circles intersect in 2 points
 */
/*****************************************************************************/
static int h3c_big_circle_intersection(_h3c_vector v1, 
                                       _h3c_vector v2,
                                       h3c_coord_t *ra,
                                       h3c_coord_t *dec) {
    _h3c_vector v;
    h3c_coord_t out[2];

    h3c_vect_prod(v1, v2, v);
    if (v[0] ==0 && v[1] == 0 && v[2] == 0) return 0;

    h3c_normalize(v);
    h3c_2radec(v, out);

    ra[0] = out[0];
    dec[0] = out[1];
    dec[1] = dec[0]*(-1);
    if (ra[0] >= 180) ra[1] = ra[0] - 180;
    else ra[1] = ra[0] + 180;
    return 1;
}

/*****************************************************************************/
/** @brief check if a point of a big circle is in a vertex 
           (defined by another big circle)
 *  @param ra right ascention
 *  @param dec declination
 *  @param ra0 right ascention of the corner of the vertex
 *  @param dec0 declination of the corner of the vertex
 *  @param ra1 right ascention of the corner of the nox
 *  @param dec1 declination of the corner of the vertex
 *  @return 1 the point is inside the box, else 0
 */
/*****************************************************************************/
static int h3c_point_in_vertex(h3c_coord_t ra, h3c_coord_t dec, 
                               h3c_coord_t ra0, h3c_coord_t dec0,
                               h3c_coord_t ra1, h3c_coord_t dec1) {
    int min, max, i;
    h3c_coord_t r;

    if (ra0 == ra1) {
        if (dec1 > dec0) { min = dec0*3600; max = dec1*3600; }
        else { min = dec1*3600; max = dec0*3600; }
        i  = dec*3600;
   
/*h3c_log(H3C_INFO," h3c_point_in_vertex min(%d)<=%d<=%d",min,i,max);*/
        if (min <= i && i <= max) return 1;
        return 0;
    }

    if (ra  < 0) ra  += 360;
    if (ra0 < 0) ra0 += 360;
    if (ra1 < 0) ra1 += 360;

    
    r = ra1-ra0; if (r < 0) r = r*(-1);
    if (r == 180) return -1; /* we don't know! */
    

    if (r > 180) {
        if (ra1 > ra0) { max = ra0*3600; min = (ra1-360)*3600; }
        else { max = ra1*3600; min = (ra0-360)*3600; }
    }
    else {
        if (ra1 > ra0) { min = ra0*3600; max = ra1*3600; }
        else { min = ra1*3600; max = ra0*3600; }
    }
     
    i  = ra*3600;

/*h3c_log(H3C_INFO," h3c_point_in_vertex min(%d)<=%d<=%d",min,i,max);*/
    if (min <= i && i <= max) return 1;
    return 0;
}

/*****************************************************************************/
/** @brief check if a polygon intersects a vertex of an other polygon 
 *  @param n1 number of vertex of the 1st polygon
 *  @param in_ra1 right ascention of the point of the 1st polygon
 *  @param in_dec1 declination of the point of the 1st polygon
 *  @param n2 number of vertex of the 1st polygon
 *  @param in_ra2 right ascention of the point of the 1st polygon
 *  @param in_dec2 declination of the point of the 1st polygon
 *  @return 1 the polygons intersects, else 0
 */
/*****************************************************************************/
int h3c_check_poly_intersect_poly(int n1,
                                  h3c_coord_t in_ra1[],
                                  h3c_coord_t in_dec1[],
                                  int n2,
                                  h3c_coord_t in_ra2[],
                                  h3c_coord_t in_dec2[])
{
    _h3c_vector *vect_prod1, vect_prod2;
    _h3c_vector va2[n2];
    h3c_coord_t ra[2], dec[2];
    int i, j, k;

    /* compute the vect. product */
    vect_prod1 = h3c_poly_vect_prod(in_ra1, in_dec1, n1);

    /* compute the vect. product */
    h3c_2xyz(in_ra2[0], in_dec2[0], va2[0]);
    for (i = 0; i < n2; i++) {
        if (i < n2-1) h3c_2xyz(in_ra2[i+1], in_dec2[i+1], va2[i+1]);
        h3c_vect_prod(va2[i], va2[(i+1)%n2], vect_prod2);
        h3c_normalize(vect_prod2);

        for (j = 0; j < n1; j++) {
            if ( !h3c_big_circle_intersection(vect_prod2, vect_prod1[j], ra, dec)) {
                /* the same circle */
                ra[0] = in_ra1[j]; ra[1] = in_ra1[(j+1)%n1];
                dec[0]  = in_dec1[j]; dec[1] = in_dec1[(j+1)%n1];
            }
/*
h3c_log(H3C_INFO, "[(%f,%f);(%f,%f)], [(%f,%f);(%f,%f)], (%f,%f);(%f,%f)",
in_ra1[j], in_dec1[j], in_ra1[(j+1)%n1], in_dec1[(j+1)%n1], 
in_ra2[i], in_dec2[i], in_ra2[(i+1)%n2], in_dec2[(i+1)%n2], 
ra[0],dec[0],ra[1],dec[1]);*/

            for (k = 0; k < 2; k++) {
#if 0
                if (h3c_check_sphere_point_in_poly_convex(n2, in_ra2, in_dec2, ra[k], dec[k]) == H3C_DISJUNCT)
                    continue;
h3c_log(H3C_INFO, "ok inside poly2");
                if (h3c_check_sphere_point_in_poly_convex(n1, in_ra1, in_dec1, ra[k], dec[k]) != H3C_DISJUNCT) {
h3c_log(H3C_INFO, "ok inside poly1");
                    free(vect_prod1);
                    return 1;
                }
#else
                if (h3c_point_in_vertex(ra[k], dec[k], 
                                        in_ra1[j], in_dec1[j],
                                        in_ra1[(j+1)%n1], in_dec1[(j+1)%n1]) != 1)
                    continue;
                if (h3c_point_in_vertex(ra[k], dec[k],
                                        in_ra2[i], in_dec2[i],
                                        in_ra2[(i+1)%n2], in_dec2[(i+1)%n2]) == 1)
                    return 1;
#endif

            }
        }
    }

    free(vect_prod1);
    return 0;
}

/*****************************************************************************/
/** @brief calculate the area of a polygon
 *  @param in_ra  right ascentions of the points of the polygon
 *  @param in_dec declinations of the points of the polygon
 *  @param nnumber of vertex of the polygon
 *  @return the area
 *
 *  @note
 *  - algo from F.Ochsenbein
 */
/*****************************************************************************/
h3c_coord_t h3c_polygon_area(h3c_coord_t *in_ra,
                             h3c_coord_t *in_dec,
                             int n) {
    int i;
    double s;
    _h3c_vector v1, v2;
    _h3c_vector *vect = (_h3c_vector *)malloc(sizeof(_h3c_vector)*n);

    /* compute the cart. coord. */
    for (i = 0; i < n; i++) h3c_2xyz(in_ra[i], in_dec[i], vect[i]);

    /* vectorial product */
    s = (2-n)*180.;
    h3c_vect_prod(vect[0], vect[n-1], v1); h3c_normalize(v1);
    h3c_vect_prod(vect[0],vect[1], v2); h3c_normalize(v2);
    s += 90-asin(h3c_scalar_product(v1, v2))*H3C_RADEG;
    for (i = 1; i < n; i++) {
        h3c_vect_prod(vect[i], vect[i-1], v1); h3c_normalize(v1);
        h3c_vect_prod(vect[i], vect[(i+1)%n], v2); h3c_normalize(v2);
        s += (90-asin(h3c_scalar_product(v1, v2))*H3C_RADEG);
    }

    free(vect);
    return s*H3C_RADEG;
}

#if H3C_BOX
/*****************************************************************************/
/** @brief set box in a polygon
    @param ra  right ascention of the center
    @param dec declination of the center
    @paral width  length in deg for the longitude
    @paral height length in deg for the lattitude
    @param in_ra  (OUT) RA coordinate of the polygon
    @param in_dec (OUT) DEC coordinate of the polygon
    @note a box is particular polygon defined by vertex in big circle and
         defining a cross centered on the center position.

*/
/*****************************************************************************/
void h3c_box_2_polygon(h3c_coord_t in_ra, h3c_coord_t in_dec,
                       h3c_coord_t width, h3c_coord_t height,
                       h3c_coord_t poly_ra[4], h3c_coord_t poly_dec[4]) {
    h3c_coord_t R[3][3];
    h3c_coord_t center[2];
    h3c_coord_t border[4][2], b[3];
    int i;

    center[0] = in_ra; center[1] = in_dec;

    /* Rotate the center to (0,0) (because problems in pole, ...) */
    tr_oR(center, R);

    border[0][0] = tand(-1*width/2.); border[0][1] = tand(-1*height/2.);
    border[1][0] = -1*border[0][0]  ; border[1][1] = border[0][1];
    border[2][0] = border[1][0]     ; border[2][1] = -1*border[0][1];
    border[3][0] = border[0][0]     ; border[3][1] = border[2][1];

    for (i=0; i<4; i++) {
        /* projection */
        tr_pu(_PROJ_TAN_, border[i], b);
        /* rotation to the original coord */
        tr_uu1(b, b, R);
        /* set in polar coordinate */
        tr_uo(b, border[i]);

        poly_ra[i]  = (border[i][0]<180) ? (h3c_coord_t) border[i][0] : (h3c_coord_t) -1*(360-border[i][0]);
        poly_dec[i] = (h3c_coord_t) border[i][1];
   }

   return;
}
#else
/* This function doesn't work with box using large circle! */
void h3c_box_2_polygon(h3c_coord_t in_ra, h3c_coord_t in_dec,
                       h3c_coord_t width, h3c_coord_t height,
                       h3c_coord_t poly_ra[4], h3c_coord_t poly_dec[4]) {
    poly_ra[0] = in_ra-width/2.; poly_dec[0] = in_dec-height/2.;
    poly_ra[1] = in_ra+width/2.; poly_dec[1] = poly_dec[0];
    poly_ra[2] = poly_ra[1]    ; poly_dec[2] = in_dec+height/2.;
    poly_ra[3] = poly_ra[0]    ; poly_dec[3] = poly_dec[2];
}
#endif
